import argparse
import numpy as np
from agt_server.agents.base_agents.bosii_agent import BOSIIAgent
from agt_server.local_games.bosii_arena import BOSIIArena


class FictitiousPlayAgent(BOSIIAgent):
    def setup(self):
        self.COMPROMISE, self.STUBBORN = 0, 1
        self.GOOD_MOOD, self.BAD_MOOD = 0, 1
        self.actions = [self.COMPROMISE, self.STUBBORN]
        self.opp_action_history = []
        self.mood_history = []

    def get_action(self):
        dist_good, dist_bad = self.predict()
        best_move = self.optimize(dist_good, dist_bad)
        return self.actions[best_move]

    def update(self):
        """
        Updates opp action history to be a record of opponent moves
        Updates mood history to be a record of the column player's historical moods
        """
        self.opp_action_history = self.get_opp_action_history()
        self.mood_history = self.get_mood_history()

    def predict(self):
        """
        Uses the opponent’s previous moves (self.opp_action_history) to generate and save a probability distribution
        over the opponent’s next move in (self.dist).
        """
        dist_good = np.zeros(len(self.actions))
        dist_bad = np.zeros(len(self.actions))
        for a, mood in zip(self.opp_action_history, self.mood_history):
            if mood == self.GOOD_MOOD:
                dist_good[a] += 1
            elif mood == self.BAD_MOOD:
                dist_bad[a] += 1
        if sum(dist_good) == 0:
            dist_goodp = np.ones(len(self.actions))/len(self.actions)
        else:
            dist_goodp = dist_good / sum(dist_good)

        if sum(dist_bad) == 0:
            dist_badp = np.ones(len(self.actions))/len(self.actions)
        else:
            dist_badp = dist_bad / sum(dist_bad)
        return dist_goodp, dist_badp

    def optimize(self, dist_good, dist_bad):
        """
        Given the distribution over the opponent's next move (output of predict) and knowledge of the payoffs (self.utility),
        Return the best move according to Ficticious Play.
        Please return one of [self.COMPROMISE, self.STUBBORN]
        """
        # TODO Calculate the expected payoff of each action and return the action with the highest payoff
        action_utils = np.zeros(len(self.actions))
        if self.is_row_player():
            for i, a in enumerate(self.actions):
                # Calculate the payoff
                for j, a in enumerate(self.actions):
                    action_utils[i] += self.col_player_good_mood_prob() * \
                        dist_good[j] * self.row_player_calculate_util(i, j)
                    action_utils[i] += (1 - self.col_player_good_mood_prob()) * \
                        dist_bad[j] * self.row_player_calculate_util(i, j)
        else:
            for i, a in enumerate(self.actions):
                # Calculate the payoff
                for j, a in enumerate(self.actions):
                    if self.get_mood() == self.GOOD_MOOD:
                        action_utils[i] += dist_good[j] * \
                            self.col_player_calculate_util(i, j, self.get_mood())
                    elif self.get_mood() == self.BAD_MOOD:
                        action_utils[i] += dist_bad[j] * \
                            self.col_player_calculate_util(i, j, self.get_mood())

        best_action = np.argmax(action_utils)
        return best_action

################### SUBMISSION #####################
agent_submission = FictitiousPlayAgent("Ficticious Play")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    agent = FictitiousPlayAgent(args.agent_name)
    if args.join_server:
        agent.connect(ip=args.ip, port=args.port)
    else:
        arena = BOSIIArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent,
                FictitiousPlayAgent("Agent_1"),
                FictitiousPlayAgent("Agent_2"),
                FictitiousPlayAgent("Agent_3"),
                FictitiousPlayAgent("Agent_4")
            ]
        )
        arena.run()



from agt_server.agents.base_agents.agent import Agent
from agt_server.utils import extract_first_json

import json
import pandas as pd
import threading
import pkg_resources


class BOSIIAgent(Agent):
    def __init__(self, name=None, timestamp=None):
        super().__init__(name, timestamp)
        self.GOOD_MOOD, self.BAD_MOOD = 0, 1
        self.mood = None
        self.utils = {self.GOOD_MOOD:
                      [[(0, 0), (3, 7)],
                       [(7, 3), (0, 0)]],
                      self.BAD_MOOD:
                      [[(0, 3), (3, 0)],
                       [(7, 0), (0, 7)]]}
        config_path = pkg_resources.resource_filename('agt_server', 'configs/server_configs/bosii_config.json')
        with open(config_path) as cfile:
            server_config = json.load(cfile)
        self.response_time = server_config['response_time']

    def timeout_handler(self):
        print(f"{self.name} has timed out")
        self.timeout = True

    def handle_permissions(self, resp):
        self.player_type = resp['player_type']
        if 'all' in resp['permissions']:
            self.game_report.game_history['my_action_history'].append(
                resp['my_action'])
            self.game_report.game_history['my_utils_history'].append(
                resp['my_utils'])
            self.game_report.game_history['opp_action_history'].append(
                resp['opp_action'])
            self.game_report.game_history['opp_utils_history'].append(
                resp['opp_utils'])
            self.game_report.game_history['mood_history'].append(
                resp['mood']
            )
        else:
            for perm in resp['permissions']:
                self.game_report.game_history[f'{perm}_history'].append(
                    resp[perm])

    def handle_postround_data(self, resp):
        self.global_timeout_count = resp['global_timeout_count']
        self.curr_opps = resp['opp_names']
        self.handle_permissions(resp)

    def play(self):
        data = self.client.recv(1024).decode()
        data = extract_first_json(data)
        
        if data:
            resp = json.loads(data)
            if resp['message'] == 'provide_game_name':
                print(f"We are playing {resp['game_name']}")
                message = {
                    "message": "game_name_recieved",
                }
                self.client.send(json.dumps(message).encode())
                self.restart()
        while True:
            data = self.client.recv(10000).decode()
            data = extract_first_json(data)
            
            if data:
                request = json.loads(data)
                if request['message'] == 'send_preround_data':
                    self.player_type = request['player_type']
                    self.mood = request['mood']
                    message = {"message": "preround_data_recieved"}
                    self.client.send(json.dumps(message).encode())
                    continue
                elif request['message'] == 'request_action':
                    self.timeout = False
                    try:
                        timer = threading.Timer(self.response_time, self.timeout_handler)
                        timer.start()
                        action = self.get_action()
                    finally:
                        if self.timeout: 
                            action = -1
                        timer.cancel()

                    try:
                        action = int(action)
                        message = {
                            "message": "provide_action",
                            "action": action, 
                            "timeout": self.timeout
                        }
                        json_m = json.dumps(message).encode()
                        self.client.send(json_m)
                    except:
                        print("Warning: Get Action must return an Integer")
                        message = {
                            "message": "provide_action",
                            "action": -1, 
                            "timeout": self.timeout
                        }
                        json_m = json.dumps(message).encode()
                        self.client.send(json_m)
                elif request['message'] == 'game_end':
                    if request['send_results']:
                        try: 
                            df = pd.read_json(request['results'])
                            if df is not None:
                                print(df)
                        except: 
                            print("Results too large. Please check in with your Lab TA for the results")
                        
                    else:
                        print(request['results'])
                    self.close()
                    break

            data = self.client.recv(10000).decode()
            data = extract_first_json(data)
            
            if data:
                resp = json.loads(data)
                if resp['message'] == 'prepare_next_game':
                    self.print_results()
                    self.restart()
                    message = {"message": "ready_next_game"}
                    self.client.send(json.dumps(message).encode())
                elif resp['message'] == 'prepare_next_round':
                    self.handle_postround_data(resp)
                    self.update()
                    message = {"message": "ready_next_round"}
                    self.client.send(json.dumps(message).encode())
                elif resp['message'] == 'disqualified':
                    if resp['disqualification_message']:
                        print(resp['disqualification_message'])
                    self.close()
                    break

    def print_results(self):
        action_counts = [0, 0, 0]
        for action in self.game_report.game_history['my_action_history']:
            if action in [0, 1]:
                action_counts[action] += 1
            else:
                action_counts[2] += 1
        print(f"Game {self.game_num}:")
        if self.curr_opps: 
            and_str = ''
            if len(self.curr_opps) > 1: 
                and_str += ', and '
            print(f"I am currently playing against {', '.join(self.curr_opps[:-1]) + and_str + self.curr_opps[-1]}")
        print(
            f"{self.name} was COMPROMISING {action_counts[0]} times and was STUBBORN {action_counts[1]} times")
        if action_counts[2] > 0:
            print(f"{self.name} submitted {action_counts[2]} invalid moves")
        if self.global_timeout_count > 0:
            print(f"{self.name} timed out {self.global_timeout_count} times")
        total_util = sum(self.game_report.game_history['my_utils_history'])
        avg_util = total_util / len(self.game_report.game_history['my_utils_history'])

        print(
            f"{self.name} got a total utility of {total_util} and a average utility of {avg_util}")
        self.game_num += 1

    def get_action_history(self):
        return self.game_report.get_action_history()

    def get_util_history(self):
        return self.game_report.get_util_history()

    def get_opp_action_history(self):
        return self.game_report.get_opp_action_history()

    def get_opp_util_history(self):
        return self.game_report.get_opp_util_history()

    def get_mood_history(self):
        return self.game_report.get_mood_history()

    def get_last_action(self):
        return self.game_report.get_last_action()

    def get_last_util(self):
        return self.game_report.get_previous_util()

    def get_opp_last_action(self):
        return self.game_report.get_opp_last_action()

    def get_opp_last_util(self):
        return self.game_report.get_opp_last_util()

    def get_last_mood(self):
        return self.game_report.get_last_mood()

    def is_row_player(self):
        return self.player_type == "row_player"

    def get_mood(self):
        return self.mood

    def row_player_calculate_util(self, row_move, col_move):
        return self.utils[self.GOOD_MOOD][row_move][col_move][0]

    def col_player_calculate_util(self, row_move, col_move, mood):
        return self.utils[mood][row_move][col_move][1]

    def col_player_good_mood_prob(self):
        return 2/3



from agt_server.agents.base_agents.lemonade_agent import LemonadeAgent
from agt_server.local_games.lemonade_arena import LemonadeArena
import argparse
import random
import numpy as np

"""
Calculates the score obtained by a move given the opponents' moves

Params:
    me - an int mod 12, the move to score
    opp_1 - an int mod 12, opponent 1's move
    opp_2 - an int mod 12, opponent 2's move
Returns:
    an int - the score obtained by the given move
"""


def calc_score(me, opp_1, opp_2):
    if me == opp_1 and me == opp_2:
        return 8
    if me == opp_1 or me == opp_2:
        return 6

    shifted_1 = (opp_1 - me) % 12
    shifted_2 = (opp_2 - me) % 12
    bigger = max(shifted_1, shifted_2)
    smaller = min(shifted_1, shifted_2)
    return 12 - bigger + smaller


"""
Given the two opponents' moves, randomly picks from the optimal responses

Params:
    opp_1 - an int mod 12, opponent 1's move
    opp_2 - an int mod 12, opponent 2's move
Returns:
    an int mod 12 - a randomly chosen optimal response
"""


def best_move(opp_1, opp_2):
    max_score = 0
    for i in range(12):
        max_score = max(max_score, calc_score(i, opp_1, opp_2))

    optimal_moves = []
    for i in range(12):
        if calc_score(i, opp_1, opp_2) == max_score:
            optimal_moves.append(i)

    index = random.randrange(len(optimal_moves))
    return optimal_moves[index]

class TeamPlayer(LemonadeAgent):

    def setup(self):
        pass

    def get_action(self):
        try:
            if random.randint(0, 1) % 2 == 0:
                opp1 = self.get_opp1_action_history()
                opp2 = self.get_opp2_action_history()
            else:
                opp1 = self.get_opp2_action_history()
                opp2 = self.get_opp1_action_history()
            # stay far away from stable ones
            if opp1[-2] == opp1[-1] and (opp1[-1] - opp2[-1]) % 12 != 6:
                action = (opp1[-1] + 6) % 12
            elif opp2[-2] == opp2[-1] and (opp1[-1] - opp2[-1]) % 12 != 6:
                action = (opp2[-1] + 6) % 12
            else:
                # action = best_move(self.opponent_1_actions[-1], self.opponent_2_actions[-1])
                biased_flip = True if random.randint(0, 3) % 4 == 0 else False
                action = random.randint(0, 11) if biased_flip else self.get_action_history()[-1]
        except:
            action = random.randint(0, 11)
        return action

    def update(self):
        pass

################### SUBMISSION #####################
agent_submission = TeamPlayer("tp")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    # parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    if args.join_server:
        agent_submission.connect(ip=args.ip, port=args.port)
    else:
        arena = LemonadeArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent_submission,
                TeamPlayer("Agent_1"),
                TeamPlayer("Agent_2"),
                TeamPlayer("Agent_3"),
                TeamPlayer("Agent_4")
            ]
        )
        arena.run()



from agt_server.agents.base_agents.cm_agent import CompleteMatrixAgent
import json
import pkg_resources

class ChickenAgent(CompleteMatrixAgent):
    def __init__(self, name=None, timestamp=None):
        super().__init__(name, timestamp)
        self.valid_actions = [0, 1]
        self.utils = [[(0, 0), (-1, 1)],
                      [(1, -1), (-5, -5)]]
        self.invalid_move_penalty = -5
        config_path = pkg_resources.resource_filename('agt_server', 'configs/server_configs/chicken_config.json')
        with open(config_path) as cfile:
            server_config = json.load(cfile)
        self.response_time = server_config['response_time']
        
    def print_results(self):
        action_counts = [0, 0, 0]
        for action in self.game_report.game_history['my_action_history']:
            if action in [0, 1]:
                action_counts[action] += 1
            else:
                action_counts[2] += 1
        print(f"Game {self.game_num}:")
        if self.curr_opps: 
            and_str = ''
            if len(self.curr_opps) > 1: 
                and_str += ', and '
            print(f"I am currently playing against {', '.join(self.curr_opps[:-1]) + and_str + self.curr_opps[-1]}")
        print(
            f"{self.name} SWERVED {action_counts[0]} times and CONTINUED {action_counts[1]} times")
        if action_counts[2] > 0:
            print(f"{self.name} submitted {action_counts[2]} invalid moves")
        if self.global_timeout_count > 0: 
            print(f"{self.name} timed out {self.global_timeout_count} times")
        total_util = sum(self.game_report.game_history['my_utils_history'])
        avg_util = total_util / len(self.game_report.game_history['my_utils_history'])

        print(
            f"{self.name} got a total utility of {total_util} and a average utility of {avg_util}")
        self.game_num += 1



import os
import pkg_resources
import importlib.util
import sys
import json
from datetime import datetime
import argparse

def import_agent_submission_from_path(path):
    directory = os.path.dirname(path)
    
    original_sys_path = sys.path.copy()
    if directory not in sys.path:
        sys.path.insert(0, directory)
    
    try:
        spec = importlib.util.spec_from_file_location("my_agent_module", path)
        my_agent = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(my_agent)
        return my_agent.agent_submission
    finally:
        sys.path = original_sys_path

def get_agent_submissions(directory):
    print(directory)
    agent_submissions = []
    folder_mod_times = {}
    for root, _, files in os.walk(directory):
        print(files)
        root_mod_time = os.path.getmtime(root)
        if root not in folder_mod_times or root_mod_time > folder_mod_times[root]:
            folder_mod_times[root] = root_mod_time
        
        for file in files:
            file_path = os.path.join(root, file)
            try:
                file_mod_time = os.path.getmtime(file_path)
                if file_mod_time > folder_mod_times[root]:
                    folder_mod_times[root] = file_mod_time
            except Exception as e:
                print(f"Failed to get modification time for {file_path}: {e}", file=sys.stderr)
        
        for file in files:
            if file == 'my_agent.py':
                full_path = os.path.join(root, file)
                try:
                    agent_submission = import_agent_submission_from_path(full_path)
                    agent_submission.timestamp = folder_mod_times[root]
                    agent_submissions.append(agent_submission)
                except Exception as e:
                    print(f"Failed to import {full_path}: {e}", file=sys.stderr)
    print(f"{len(agent_submissions)} sucessfully imported")
    return agent_submissions

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Handin Script')
    parser.add_argument('handin_config', type=str, help='The name of the handin_config file that you want to use')
    parser.add_argument('agent_submissions', type=str, help='The path to the directory that you want to run the handin-script on (Relative to agt_server)')
    args = parser.parse_args()
    full_config = f"configs/handin_configs/{args.handin_config}"
    
    config_path = pkg_resources.resource_filename('agt_server', full_config)
    with open(config_path) as cfile:
        server_config = json.load(cfile)
    
    agent_submissions = get_agent_submissions(pkg_resources.resource_filename('agt_server', args.agent_submissions))
    print(agent_submissions)
    
    directory = os.path.dirname(pkg_resources.resource_filename('agt_server', server_config['arena_path']))
    spec = importlib.util.spec_from_file_location("module.name", pkg_resources.resource_filename('agt_server', server_config['arena_path']))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    
    start = datetime.now()
    print([agent.name for agent in agent_submissions if agent is not None])
    arena_type = getattr(module, server_config['arena_classname'])
    arena = arena_type(
            num_rounds = server_config['num_rounds'],
            timeout = server_config['response_time'],
            players = [agent for agent in agent_submissions if agent is not None],
            handin = True, 
            logging_path = server_config['logging_path'], 
            summary_path = server_config['save_path']
        )
    arena.run()
    end = datetime.now()
    print(f"Time Elapsed (s): {(end - start).total_seconds()}")
    
    


import random
import numpy as np
import pandas as pd
from collections import defaultdict
from agt_server.local_games.base import LocalArena
from itertools import permutations

class GSVM9Arena(LocalArena):
    """
    A local arena for the GSVM-9 Auction with:
      - 4 Bidders: [National, Regional1, Regional2, Regional3]
      - 9 Goods: A..I
      - Uniform(0, maxVal) valuations from the table
      - Global synergy: (1 + 0.2*(n-1)) * sum_of_valuations
      - Capacity: National can win up to 6 goods, Regionals up to 3 each
    """

    def __init__(self,
                 players=[],
                 num_rounds=10,
                 timeout=1,
                 handin=False,
                 logging_path=None,
                 save_path=None):
        """
        :param players: List of exactly 4 players [N, R1, R2, R3].
        :param num_rounds: Number of rounds in the repeated game.
        :param timeout: Timeout in seconds for get_action calls.
        :param handin: Boolean indicating whether to suppress prints and store logs in a file.
        :param logging_path: File path for logging if handin=True.
        :param save_path: Path for final results if handin=True.
        """
        super().__init__(num_rounds, players, timeout, handin, logging_path, save_path)

        # We expect exactly 4 players: 1 national, 3 regional.
        assert len(self.players) >= 4, "GSVM-9 requires at least 4 players."

        # Define the set of goods
        self.goods = ["A", "B", "C", "D", "E", "F", "G", "H", "I"]
        self.game_name = "GSVM-9 Auction"

        # Max valuations from the table (blank entries => 0 => not eligible)
        # By default, assume players[0] = National, players[1..3] = Regionals
        self.max_valuations = [
            # National Bidder
            {"A":15, "B":15, "C":30, "D":30, "E":15, "F":15, "G":0,  "H":0,  "I":0},
            # Regional Bidder 1
            {"A":20, "B":20, "C":40, "D":40, "E":20, "F":0,  "G":20, "H":0,  "I":0},
            # Regional Bidder 2
            {"A":0,  "B":40, "C":40, "D":20, "E":20, "F":0,  "G":0,  "H":20, "I":0},
            # Regional Bidder 3
            {"A":20, "B":20, "C":0,  "D":20, "E":20, "F":20, "G":0,  "H":0,  "I":20}
        ]

        # Capacity: National = 6, Regionals = 3
        self.capacity = [6, 3, 3, 3]

        # Initialize game reports
        for idx, player in enumerate(self.players):
            self.game_reports[player.name] = {
                "valuation_history": [],
                "bid_history": [],
                "util_history": [],
                "index": idx,
                "global_timeout_count": 0
            }

        self.results = []
        self.game_num = 1

    def run(self):
        """
        Run the entire repeated auction for self.num_rounds rounds.
        """
        # Restart all agents first
        for player in self.players:
            self.run_func_w_time(player.restart, self.timeout, player.name)
        
        for group in permutations(self.players, 4):
            self.run_game(group)

        # Summarize final results
        results = self.summarize_results()
        return results

    def run_game(self, group):
        """
        Run all rounds of the auction among these 4 players.
        """
        self._log_or_print(f"\n{group[0].name} (National) vs. {group[1].name} (R1) vs. {group[2].name} (R2) vs. {group[3].name} (R3)")
        group[0]._is_national_bidder = True
        group[1]._is_national_bidder = False
        group[2]._is_national_bidder = False
        group[3]._is_national_bidder = False

        for _ in range(self.num_rounds):
            # 1) Assign valuations
            for i, player in enumerate(group):
                valuations = {}
                for good in self.goods:
                    max_val = self.max_valuations[i].get(good, 0)
                    if max_val > 0:
                        valuations[good] = random.uniform(0, max_val)
                    else:
                        valuations[good] = 0.0  # ineligible => 0
                # Store the valuations in the agent
                player.valuations = valuations
                # Save in game_reports
                self.game_reports[player.name]["valuation_history"].append(valuations)

            # 2) Gather bids
            bids = {}
            for player in group:
                bid_dict = self.run_func_w_time(player.get_action, self.timeout, player.name, {})
                if not isinstance(bid_dict, dict):
                    bid_dict = {}
                bids[player.name] = bid_dict
                self.game_reports[player.name]["bid_history"].append(bid_dict)

            # 3) Compute outcome (respect capacity)
            allocation, payments = self.compute_auction_result(bids, players=group)

            # 4) Compute synergy-based utility
            for i, player in enumerate(group):
                # Identify which goods were won
                won_goods = [g for g, w in allocation.items() if w == player.name]
                # Sum of valuations for these goods
                base_sum = sum(player.valuations.get(g, 0) for g in won_goods)
                n = len(won_goods)
                synergy_val = (1 + 0.2*(n-1)) * base_sum if n > 0 else 0
                cost = payments[player.name]
                util = synergy_val - cost

                self.game_reports[player.name]["util_history"].append(util)
                # Optionally store in the player's game_report object:
                if hasattr(player, 'game_report'):
                    if "my_utils_history" not in player.game_report.game_history:
                        player.game_report.game_history["my_utils_history"] = []
                    player.game_report.game_history["my_utils_history"].append(util)

            # 5) Let agents update
            for player in group:
                self.run_func_w_time(player.update, self.timeout, player.name)

            self.game_num += 1

        # After finishing all rounds, print each player's total utility
        self._log_or_print(f"Winners: {allocation}")
        self._log_or_print(f"Prices: {payments}")
        for player in group:
            tot = sum(self.game_reports[player.name]["util_history"])
            self._log_or_print(f"{player.name} final total utility: {tot}")

    def compute_auction_result(self, bids, players):
        """
        For each good, assign it to the highest bidder that hasn't exceeded capacity
        and is eligible to bid on that good (i.e., might have a positive max_valuation).
        Returns (allocation, payments) just like in SAArena.
        """
        allocation = {}
        payments = {p.name: 0 for p in players}
        # Track how many goods each player has won so far in this round
        capacity_used = defaultdict(int)

        for good in self.goods:
            best_bid = None
            best_player = None
            best_player_idx = None
            for i, player in enumerate(players):
                # skip if capacity used is already at max
                if capacity_used[player.name] >= self.capacity[i]:
                    continue
                # skip if ineligible (max_valuation is 0 => can't place positive bid)
                if self.max_valuations[i].get(good, 0) <= 0:
                    continue

                # actual bid
                val = bids[player.name].get(good, 0) or 0
                if (best_bid is None) or (val > best_bid):
                    best_bid = val
                    best_player = player.name
                    best_player_idx = i

            allocation[good] = best_player
            if best_player is not None and best_bid is not None and best_bid > 0:
                payments[best_player] += best_bid
                capacity_used[best_player] += 1

        return allocation, payments

    def summarize_results(self):
        """
        Summarize results across all rounds:
        - total utility
        - average utility per round
        """
        summary_data = []
        for player in self.players:
            name = player.name
            util_list = self.game_reports[name]["util_history"]
            total_util = sum(util_list)
            avg_util = total_util / len(util_list) if util_list else 0
            summary_data.append([name, total_util, avg_util])

        df = pd.DataFrame(summary_data, columns=["Player", "Total Utility", "Avg Utility"])
        df = df.sort_values("Total Utility", ascending=False)
        self._log_or_print(f"\nFinal GSVM-9 Auction Results:\n{df}\n")
        return df



import json
import random
import numpy as np
import pkg_resources
import threading
from agt_server.agents.base_agents.agent import Agent
from agt_server.agents.base_agents.game_report import GameReport
from agt_server.utils import extract_first_json
import pandas as pd

class GSVM9Agent(Agent):
    """
    A minimal example agent for a GSVM-9 style server-based auction.
    Tracks whether it's a National or Regional bidder, capacity, valuations, etc.
    """

    def __init__(self, name=None, timestamp=None):
        super().__init__(name, timestamp)
        # Load config if you want (e.g., response time)
        config_path = pkg_resources.resource_filename('agt_server', 'configs/server_configs/sa_config.json')
        with open(config_path) as cfile:
            self.config = json.load(cfile)
        self.response_time = self.config.get('response_time', 2)

        # Auction-related fields
        self.num_goods = None
        self.goods = None
        self.valuations = {}
        self._goods_to_index = {}
        self._index_to_goods = {}

        # Role & capacity
        self._is_national_bidder = False
        self.capacity = 3  # default, overwritten if national

        # Timeout tracking
        self.timeout = False
        self.global_timeout_count = 0
        self.round_number = 0

    def timeout_handler(self):
        print(f"{self.name} has timed out")
        self.timeout = True

    def handle_permissions(self, resp):
        """
        If the server sends post-round data about what we did, payments, etc.
        you can store it in game_report.
        """
        self.player_type = resp.get('player_type')
        gh = self.game_report.game_history

        # Example usage: store the last round’s data
        gh.setdefault('my_bid_history', []).append(resp.get('my_bid'))
        gh.setdefault('my_utils_history', []).append(resp.get('my_util'))
        gh.setdefault('my_payment_history', []).append(resp.get('payments'))
        gh.setdefault('price_history', []).append(resp.get('prices'))
        gh.setdefault('winner_history', []).append(resp.get('winners'))

    def handle_postround_data(self, resp):
        self.global_timeout_count = resp.get('global_timeout_count', 0)
        self.handle_permissions(resp)

    def is_national_bidder(self): 
        return self._is_national_bidder
    
    def get_bids(self):
        """
        Produce a dictionary: {good_name: bid_value, ...}.
        Here we do a naive approach: we bid our entire valuation on each good
        we are eligible for. If we want to ensure we don't exceed capacity
        (for example, 6 for national, 3 for regional), we could do something more
        sophisticated, but typically the server enforces capacity anyway.
        """
        bid_dict = {}
        # Example: bid entire valuation on each good with valuation > 0
        for good, val in self.valuations.items():
            if val > 0:
                bid_dict[good] = val
            else:
                bid_dict[good] = 0
        return bid_dict

    def get_action(self):
        """
        Called by the server to get your bid dictionary each round.
        """
        return self.get_bids()

    def print_results(self):
        """
        Prints a summary of the current game results based on the game history.
        """
        gh = self.game_report.game_history

        last_bid = gh.get('my_bid_history', [])[-1] if gh.get('my_bid_history') else None
        last_winners = gh.get('winner_history', [])[-1] if gh.get('winner_history') else None
        last_prices = gh.get('price_history', [])[-1] if gh.get('price_history') else None
        last_payment = gh.get('my_payment_history', [])[-1] if gh.get('my_payment_history') else None
        last_util = gh.get('my_utils_history', [])[-1] if gh.get('my_utils_history') else None

        self.round_number += 1
        print(f"----- Round {self.round_number} Results for {self.name} -----")
        print(f"Bidder Type: {'National' if self.is_national_bidder else 'Regional'}")
        print(f"Last Bid: {last_bid}")
        print(f"Winners: {last_winners}")
        print(f"Prices: {last_prices}")
        print(f"My Payment: {last_payment}")
        print(f"My Utility This Round: {last_util}")
        print("-----------------------------------------\n")

    def play(self):
        """
        Main loop for interacting with the GSVM-9 (or similar) server game.
        Listens for messages and responds appropriately.
        """
        # 1. Possibly read initial handshake
        data = self.client.recv(1024).decode()
        data = extract_first_json(data)
        if data:
            resp = json.loads(data)
            if resp.get('message') == 'provide_game_name':
                print(f"We are playing {resp.get('game_name')}")
                message = {"message": "game_name_recieved"}
                self.client.send(json.dumps(message).encode())
                self.restart()

        # 2. Main loop
        while True:
            data = self.client.recv(10000).decode()
            data = extract_first_json(data)
            if not data:
                continue
            request = json.loads(data)

            msg_type = request.get('message')
            if msg_type == 'send_preround_data':
                # The server tells us whether we are national or regional
                # in addition to valuations, etc.
                self.is_national_bidder = request.get('is_national', False)
                # capacity = 6 if national, else 3
                self.capacity = 6 if self.is_national_bidder else 3

                self.num_goods = request.get('num_goods')
                goods_list = request.get('goods', [])
                self.goods = set(goods_list)

                # Build index mappings
                self._goods_to_index = {good: idx for idx, good in enumerate(goods_list)}
                self._index_to_goods = {idx: good for good, idx in self._goods_to_index.items()}

                # Retrieve valuations from the server
                valuations_dict = request.get('valuations', {})
                self.valuations = {}
                for good in goods_list:
                    self.valuations[good] = valuations_dict.get(good, 0.0)

                # Acknowledge
                self.client.send(json.dumps({"message": "preround_data_recieved"}).encode())

            elif msg_type == 'request_bid':
                # The server wants our bid dictionary
                self.timeout = False
                try:
                    timer = threading.Timer(self.response_time, self.timeout_handler)
                    timer.start()
                    bid = self.get_action()
                finally:
                    if self.timeout:
                        bid = None
                    timer.cancel()

                try:
                    self.client.send(json.dumps({
                        "message": "provide_bid",
                        "bid": bid,
                        "timeout": self.timeout
                    }).encode())
                except Exception as e:
                    print(f"Error sending bid: {e}")

            elif msg_type == 'prepare_next_round':
                # The server is telling us the round is over
                self.print_results()
                self.handle_postround_data(request)
                self.update()
                self.client.send(json.dumps({"message": "ready_next_round"}).encode())

            elif msg_type == 'prepare_next_game':
                # The server is telling us the game ended and a new one will begin
                self.print_results()
                self.restart()
                self.client.send(json.dumps({"message": "ready_next_game"}).encode())

            elif msg_type == 'game_end':
                # The entire game is finished
                if request.get('send_results'):
                    try:
                        df = pd.read_json(request['results'])
                        if df is not None:
                            print(df)
                    except:
                        print("Results too large. Please check with your TA.")
                else:
                    print(request.get('results'))
                self.close()
                break

            elif msg_type == 'disqualified':
                # We were disqualified
                dq_msg = request.get('disqualification_message', '')
                if dq_msg:
                    print(dq_msg)
                self.close()
                break



from i_fixed_policy import IFixedPolicy
import numpy as np


class UniformPolicy(IFixedPolicy):
    def __init__(self, action_state_sz) -> None:
        super().__init__()
        self.action_state_sz = action_state_sz

    def get_move(self, state):
        return np.random.randint(0, self.action_state_sz)



from agt_server.agents.base_agents.agent import Agent
import json
import pandas as pd
import threading
import pkg_resources
import numpy as np
import random
from itertools import product

class MyLSVMAgent(Agent):
    def __init__(self, name=None, timestamp=None, config_path = 'configs/handin_configs/lsvm_config.json'):
        super().__init__(name, timestamp)
        
        config_path = pkg_resources.resource_filename('agt_server', config_path)
        with open(config_path) as cfile:
            server_config = json.load(cfile)
            
        self.response_time = server_config['response_time']
        self.shape = tuple(server_config['shape']) 
        self.num_goods = np.prod(self.shape)
        self._is_national_bidder = None
        self.valuations = None
        self.min_bids = np.zeros(self.shape)
        self.current_prices = np.zeros(self.shape)
        self.regional_good = None
        self._regional_size = int(server_config['regional_size'])
       
        self._goods_to_index = MyLSVMAgent._name_goods(self.shape)
        self.goods = set(self._goods_to_index.keys())
        self._index_to_goods = {value: key for key, value in self._goods_to_index.items()}
        self._current_round = 0 
        
        self.tentative_allocation = set()
    
    def get_regional_good(self): 
        """
        Return the regional good assigned to the agent.

        Returns:
        - regional_good: The specific good designated as the regional good for this agent.
        """
        return self.regional_good
    
    def get_goods(self): 
        """
        Get the set of goods names available in the auction.

        Returns:
        - A set of strings representing the names of the goods.
        """
        return self.goods 
    
    def is_national_bidder(self): 
        """
        Check if the agent is designated as a national bidder.

        Returns:
        - True if the agent is a national bidder, False otherwise.
        """
        return self._is_national_bidder
    
    def get_shape(self): 
        """
        Get the shape of the goods space as a tuple.

        Returns:
        - A tuple representing the dimensions of the goods space.
        """
        return self.shape 
    
    def get_num_goods(self): 
        """
        Get the total number of goods available.

        Returns:
        - An integer representing the total number of goods.
        """
        return self.num_goods
        
    def get_goods_to_index(self): 
        """
        Get the mapping from goods names to their index in the goods space.

        Returns:
        - A dictionary mapping string names to tuple indices.
        """
        return self._goods_to_index
    
    def get_tentative_allocation(self): 
        """
        Get the current tentative allocation of goods to this agent.

        Returns:
        - A set of goods (strings) currently tentatively allocated to this agent.
        """
        return self.tentative_allocation

    def get_partitions(self, bundle = None): 
        """
        Calculate the partitions of a bundle for the regional or national bidder.
        
        :param bundle: A set of strings, where each strings represents the name of a good
        :return: A list of connected partitions of the bundle of goods
        """
        def _is_adjacent(item1, item2):
            i1_idx = self._goods_to_index[item1]
            i2_idx = self._goods_to_index[item2]
            return sum([abs(i1_idx[i] - i2_idx[i]) for i in range(len(i1_idx))]) == 1

        def _dfs(current, visited, component, all_goods):
            visited.add(current)
            component.add(current)
            for neighbor in all_goods:
                if neighbor not in visited and _is_adjacent(current, neighbor):
                    _dfs(neighbor, visited, component, all_goods)

        def _get_partitions(all_goods):
            visited = set()
            partitions = []
            for good in all_goods:
                if good not in visited:
                    component = set()
                    _dfs(good, visited, component, all_goods)
                    partitions.append(component)
            return partitions
    
        partitions = _get_partitions(list(bundle))
        return partitions
    
    def calc_total_valuation(self, bundle = None):
        """
        Calculate the valuation of a bundle for the regional or national bidder.
        
        :param bundle: A set of strings, where each strings represents the name of a good
        :return: The valuation of the bundle.
        """
        if bundle is None: 
            bundle = self.tentative_allocation
            
        if self._is_national_bidder:
            a = 320
            b = 10
        else:
            a = 160
            b = 4
        
        base_values = {good: self.valuations[self._goods_to_index[good]] for good in bundle}
        partitions = self.get_partitions(bundle)
        
        valuation = 0
        for C in partitions:
            partition_valuation = sum(base_values[idx] for idx in C)
            valuation += (1 + a / (100 * (1 + np.exp(b - len(C))))) * partition_valuation
        return valuation
    
    def calc_total_prices(self, bundle = None):
        """
        Calculate the prices of a bundle for the regional or national bidder.
        
        :param bundle: A set of strings, where each strings represents the name of a good
        :return: The prices of the bundle.
        """
        if bundle is None: 
            bundle = self.tentative_allocation
            
        return sum([self.current_prices[self._goods_to_index[good]] for good in bundle])
         
        
    def calc_total_utility(self, bundle = None):
        """
        Calculate the utility of a bundle for the regional or national bidder.
        
        :param bundle: A set of strings, where each strings represents the name of a good
        :return: The utility of the bundle.
        """
        valuation = self.calc_total_valuation(bundle)
        prices = self.calc_total_prices(bundle) 
        return valuation - prices        

    def calculate_tentative_valuation(self): 
        """
        Calculates the total bundle valuation of the tentative allocation.
        
        Returns:
        - float: The calculated utility value.
        """
        return self.calc_bundle_valuation(self.tentative_allocation)
        
    def get_current_round(self): 
        """
        Retrieves the current round number.
        
        Returns:
        - int: The current round number.
        """
        return self._current_round
        
    def get_goods_in_proximity(self): 
        """
        Retrieves the names of goods that are within the agent's proximity.
        Includes all goods if the agent is a National Bidder.
        
        Returns:
        - list: A list of goods' names within the agent's proximity.
        """
        if self._is_national_bidder:
            return list(self._index_to_goods.values())
        else:
            non_zero_indices = np.argwhere(self.valuations != 0)
            non_zero_indices_tuples = [tuple(idx) for idx in non_zero_indices]
            return [self._index_to_goods[idx] for idx in non_zero_indices_tuples if idx in self._index_to_goods]
        
    
    def proximity(self, arr = None, regional_good = None):
        """
        Filters the valuation array to only include valuations within a specified distance from a regional good.
        
        Parameters:
        - arr (numpy.ndarray, optional): The valuation array to filter. If None, uses the agent's valuations.
        - regional_good (str, optional): The regional good used as the center for filtering. If None, uses the agent's regional good.
        
        Returns:
        - numpy.ndarray: A masked array with valuations outside the specified distance set to 0.
        """
        if arr is None: 
            arr = self.valuations
        if regional_good is None: 
            regional_good = self.regional_good
        
        index = self._goods_to_index[regional_good]
        grid = np.ogrid[tuple(slice(0, max_shape) for max_shape in arr.shape)]
        distance = sum(np.abs(g - idx) for g, idx in zip(grid, index))
        masked_arr = np.where(distance <= self._regional_size, arr, 0)
        return masked_arr
    
    @staticmethod
    def _generate_sequence(alphabet = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'):
        """
        Generates an infinite sequence of strings based on the provided alphabet.
        
        Parameters:
        - alphabet (str, optional): The alphabet to use for generating the sequence. Defaults to 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'.
        
        Yields:
        - str: The next string in the sequence.
        """
        yield from alphabet 
        size = 2 
        while True:
            for letters in product(alphabet, repeat=size):
                yield ''.join(letters)
            size += 1
        
    @staticmethod
    def _name_goods(shape, alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"):
        """
        Names the goods in a multidimensional world of a given shape using an alphabet.
        
        Parameters:
        - shape (tuple): The shape of the world.
        - alphabet (str, optional): The alphabet used for naming goods.
        
        Returns:
        - dict: A mapping from string names to multidimensional indices.
        """
        alphabet_generator = MyLSVMAgent._generate_sequence(alphabet)
        map_dict = {}
        total_elements = np.prod(shape)
        
        for flat_index in range(total_elements):
            multidimensional_index = np.unravel_index(flat_index, shape)
            letter = next(alphabet_generator)
            map_dict[letter] = multidimensional_index
                        
        return map_dict

    def get_valuation_as_array(self): 
        """
        Retrieves the agent's valuation as a numpy array.
        
        Returns:
        - numpy.ndarray: The valuation array.
        """
        return self.valuations
    
    def get_valuation(self, good): 
        """
        Retrieves the valuation for a specific good.
        
        Parameters:
        - good (str): The name of the good.
        
        Returns:
        - float: The valuation for the specified good.
        """
        return self.valuations[self._goods_to_index[good]]
    
    def get_valuations(self, bundle = None): 
        """
        Retrieves the valuations for a set of goods.
        
        Parameters:
        - bundle (set, optional): A set of goods for which valuations are retrieved. If None, uses all goods.
        
        Returns:
        - dict: A mapping from goods to their valuations.
        """
        if bundle is None: 
            bundle = self.goods
        return {good: self.valuations[self._goods_to_index[good]] for good in bundle}
    
    def get_min_bids_as_array(self): 
        """
        Retrieves the minimum bids as a numpy array.
        
        Returns:
        - numpy.ndarray: The array of minimum bids.
        """
        return self.min_bids 
    
    def get_min_bids(self, bundle = None): 
        """
        Retrieves the minimum bids for a set of goods.
        
        Parameters:
        - bundle (set, optional): A set of goods for which minimum bids are retrieved. If None, uses all goods.
        
        Returns:
        - dict: A mapping from goods to their minimum bids.
        """
        if bundle is None: 
            bundle = self.goods
        return {good: self.min_bids[self._goods_to_index[good]] for good in bundle if good in self._goods_to_index}

    def is_valid_bid_bundle(self, my_bids):
        """
        Checks whether a bundle of bids is valid according to the agent's constraints and the game's rules.
        
        Parameters:
        - my_bids (dict): A dictionary mapping goods to bid values.
        
        Returns:
        - bool: True if the bid bundle is valid, False otherwise.
        """
        
        if not isinstance(my_bids, dict):
            print("NOT VALID: my_bids must be of type Dict[str, float]")
            return False
        
        for good, bid in my_bids.items():
            if bid is None:
                print(f"NOT VALID: bid for good {good} cannot be None")
                return False

            if good not in self._goods_to_index or bid < self.min_bids[self._goods_to_index[good]]:
                print(f"NOT VALID: bid for good {good} cannot be less than the min bid")
                return False

            price_history = self.game_report.game_history['price_history']
            bid_history = self.game_report.game_history['my_bid_history']
            for past_prices, past_bids in zip(price_history, bid_history):
                price_diff = self.current_prices - past_prices
                bid_diff = my_bids - past_bids
                switch_cost = np.dot(price_diff, bid_diff)
                if switch_cost > 0:
                    print(f"NOT VALID: New bids are {switch_cost} relatively more expensive than maintaining the old bids, strategy is insincere")
                    return False  
        return True 
        
    def clip_bids(self, my_bids):
        """
        Ensures that all bids in a bid bundle are at least as high as the minimum bids for the corresponding goods.

        Parameters:
        - my_bids (dict): A dictionary mapping goods to bid values.

        Returns:
        - dict: A dictionary with the bid values adjusted to be no lower than the minimum bid requirements.
        """
        for good in my_bids: 
            my_bids[good] = max(my_bids[good], self.min_bids[self._goods_to_index[good]])
        return my_bids

    def clip_bid(self, good, bid): 
        """
        Clips a single bid for a good to ensure it meets or exceeds the minimum bid.

        Parameters:
        - good (str): The good for which the bid is being placed.
        - bid (float): The bid value for the good.

        Returns:
        - float: The adjusted bid value, which will be no lower than the minimum required bid for the specified good.
        """
        return max(bid, self.min_bids[self._goods_to_index[good]])
    
    def timeout_handler(self):
        """
        Handles timeout situations by setting the agent's timeout status and potentially performing other necessary actions.
        """
        print(f"{self.name} has timed out")
        self.timeout = True

    def handle_postround_data(self, resp):
        """
        Processes post-round data received from the server, updating the agent's state accordingly.

        Parameters:
        - resp (dict): A dictionary containing post-round data from the server.
        """
        self.global_timeout_count = resp['global_timeout_count']
        self.curr_opps = resp['opp_names']
        self.handle_permissions(resp)

    def map_to_ndarray(self, map, object = False): 
        """
        Converts a mapping of goods to values into a numpy array with the agent's valuation shape.

        Parameters:
        - map (dict): A dictionary mapping goods to values.
        - object (bool, optional): Whether to create an array of objects. Defaults to False.

        Returns:
        - numpy.ndarray: An array representing the mapping, in the shape of the agent's valuation.
        """
        if object: 
            arr = np.empty(self.shape, dtype=object)
        else: 
            arr = np.zeros(self.shape)
        for item in map: 
            arr[self._goods_to_index[item]] = map[item]
        return arr

    def ndarray_to_map(self, arr): 
        """
        Converts a numpy array into a dictionary mapping goods to values, based on the agent's goods' indexing.

        Parameters:
        - arr (numpy.ndarray): The array to convert.

        Returns:
        - dict: A dictionary mapping goods to their values in the array.
        """
        return {good: arr[self._goods_to_index[good]] for good in self.get_goods()}
                        
    def get_game_report(self): 
        """
        Retrieves the game report containing the history and outcomes of all the rounds the agent has participated in.

        Returns:
        - GameReport: The game report object.
        """
        return self.game_report

    def get_util_history(self):
        """
        Retrieves the utility history for the agent, showing how the agent's utility has changed over time.

        Returns:
        - list: A list of utility values, one for each round.
        """
        return self.game_report.get_util_history()

    def get_bid_history(self): 
        """
        Retrieves the history of bids made during the game [AS A NDARRAY].

        :return: A list of bids (np.ndarrays) if available; otherwise, an empty list.
        """
        return self.game_report.get_bid_history()
    
    def get_bid_history_map(self): 
        """
        Retrieves the history of bids made during the game [AS A MAP].

        :return: A list of bid (maps from goods to value) if available; otherwise, an empty list.
        """
        return self.game_report.get_bid_history_map()
    
    def get_price_history(self): 
        """
        Retrieves the history of prices during the game [AS A NDARRAY].

        :return: A list of prices (np.ndarrays) if available; otherwise, an empty list.
        """
        return self.game_report.get_price_history()
    
    def get_price_history_map(self): 
        """
        Retrieves the history of prices during the game [AS A MAP].

        :return: A list of prices (maps from goods to value) if available; otherwise, an empty list.
        """
        return self.game_report.get_price_history_map()
    
    def get_winner_history(self): 
        """
        Retrieves the history of tentative winners during the game [AS A NDARRAY of OBJECTS].

        :return: A list of winners (np.ndarrays) if available; otherwise, an empty list.
        """
        return self.game_report.get_winner_history()
    
    def get_winner_history_map(self): 
        """
        Retrieves the history of tentative winners during the game [AS A MAP].

        :return: A list of winners (maps from goods to winner) if available; otherwise, an empty list.
        """
        return self.game_report.get_winner_history_map() 
    
    def get_previous_util(self):
        """
        Retrieves the most recent utility value for the agent.

        Returns:
        - float: The last utility value recorded for the agent.
        """
        return self.game_report.get_previous_util()
    
    def get_previous_bid(self):
        """
        Retrieves the most recent entry from the bid history as a ndarray. 
        :return: The latest bids if available; otherwise, None.
        """
        return self.game_report.get_previous_bid()

    def get_previous_bid_map(self):
        """
        Retrieves the most recent entry from the bid history as a map. 
        :return: The latest bids if available; otherwise, None.
        """
        return self.game_report.get_previous_bid_map()
    
    def get_current_prices(self): 
        """
        Retrieves the most recent entry from the price history as a ndarray. 
        :return: The latest price if available; otherwise, None.
        """
        return self.game_report.get_previous_price()
    
    def get_current_prices_map(self): 
        """
        Retrieves the most recent entry from the price history as a map. 
        :return: The latest price if available; otherwise, None.
        """
        return self.game_report.get_previous_price_map()
    
    def get_previous_winners(self): 
        """
        Retrieves the most recent entry from the tentative winners history as a ndarray. 
        :return: The latest winners if available; otherwise, None.
        """
        return self.game_report.get_previous_winners()
    
    def get_previous_winners_map(self): 
        """
        Retrieves the most recent entry from the tentative winners history as a map. 
        :return: The latest winners if available; otherwise, None.
        """
        return self.game_report.get_previous_winners_map()
    


from agt_server.local_games.base import LocalArena
from itertools import combinations
import numpy as np
import pandas as pd
import os
import json
from datetime import datetime


class RPSArena(LocalArena):
    def __init__(self, num_rounds=10, players=[], timeout=1, handin=False, 
                 logging_path=None, summary_path=None, detailed_reports_path=None):
        super().__init__(num_rounds, players, timeout, handin, logging_path, summary_path)
        self.game_name = "Rock, Paper, Scissors"
        self.valid_actions = [0, 1, 2]
        self.action_map = {0: "Rock", 1: "Paper", 2: "Scissors"}  # Map numbers to actions
        self.utils = [[0, -1, 1],  # Payoff matrix
                      [1, 0, -1],
                      [-1, 1, 0]]
        self.invalid_move_penalty = -1
        self.detailed_reports_path = detailed_reports_path

        if not self.handin_mode:
            assert len(self.players) >= 2, "Arena must have at least 2 players"

        for idx, player in enumerate(self.players):
            self.game_reports[player.name] = {
                "action_history": [],
                "util_history": [],
                "index": idx,
                "timeout_count": 0,
                "global_timeout_count": 0,
                "disconnected": False
            }

        self.result_table = np.zeros((len(players), len(players)))
        self.final_utilities = {player.name: 0 for player in self.players}
        self.game_num = 1

    def calculate_utils(self, p1_action, p2_action):
        if p1_action not in self.valid_actions and p2_action not in self.valid_actions:
            return [0, 0]
        if p1_action not in self.valid_actions:
            return [self.invalid_move_penalty, 0]
        if p2_action not in self.valid_actions:
            return [0, self.invalid_move_penalty]
        return [self.utils[p1_action][p2_action], self.utils[p2_action][p1_action]]

    def reset_game_reports(self):
        for player in self.players:
            self.game_reports[player.name]["action_history"] = []
            self.game_reports[player.name]["util_history"] = []

    def run_game(self, p1, p2):
        p1_total_util = 0
        p2_total_util = 0

        for _ in range(self.num_rounds):
            p1_action = self.run_func_w_time(p1.get_action, self.timeout, p1.name, -1)
            p2_action = self.run_func_w_time(p2.get_action, self.timeout, p2.name, -1)

            
            self.game_reports[p1.name]['action_history'].append(p1_action)
            self.game_reports[p2.name]['action_history'].append(p2_action)
            p1.game_report.game_history['my_action_history'].append(p1_action)
            p2.game_report.game_history['my_action_history'].append(p2_action)
            p1.game_report.game_history['opp_action_history'].append(p2_action)
            p2.game_report.game_history['opp_action_history'].append(p1_action)


            p1_util, p2_util = self.calculate_utils(p1_action, p2_action)
            self.game_reports[p1.name]['util_history'].append(p1_util)
            self.game_reports[p2.name]['util_history'].append(p2_util)
            p1.game_report.game_history['my_utils_history'].append(p1_util)
            p2.game_report.game_history['my_utils_history'].append(p2_util)
            p1.game_report.game_history['opp_utils_history'].append(p2_util)
            p2.game_report.game_history['opp_utils_history'].append(p1_util)

            p1_total_util += p1_util
            p2_total_util += p2_util
            self.run_func_w_time(p1.update, self.timeout, p1.name, -1)
            self.run_func_w_time(p2.update, self.timeout, p1.name, -1)

        # Update results table
        p1_index = self.game_reports[p1.name]["index"]
        p2_index = self.game_reports[p2.name]["index"]
        self.result_table[p1_index, p2_index] += p1_total_util
        self.result_table[p2_index, p1_index] += p2_total_util

        # Update final utilities
        self.final_utilities[p1.name] += p1_total_util
        self.final_utilities[p2.name] += p2_total_util

        # Summarize the game
        game_summary = self._generate_game_summary(p1, p2)
        self._log_or_print(game_summary)

        # Save detailed game report if needed
        if self.detailed_reports_path:
            self._save_detailed_game_report(p1, p2)

    def run(self):
        for p1, p2 in combinations(self.players, 2):
            self.run_func_w_time(p1.restart, self.timeout, p1.name)
            self.run_func_w_time(p2.restart, self.timeout, p2.name)
            self.run_game(p1, p2)
            self.game_num += 1

        return self.summarize_results()

    def summarize_results(self):
        agent_names = [player.name for player in self.players]
        df = pd.DataFrame(self.result_table, columns=agent_names, index=agent_names)

        # Add final and average utility columns
        df["Final Utility"] = [self.final_utilities[player.name] for player in self.players]
        df["Average Utility"] = df["Final Utility"] / (len(self.players) - 1)  # Average across opponents

        # Log or save results
        if self.save_path:
            df.to_csv(self.save_path)

        self._log_or_print("\nFinal Results:\n")
        self._log_or_print(df.to_string())

        return df

    def _generate_game_summary(self, p1, p2):
        """
        Generate a summary for a specific game between two players.
        """
        p1_actions = self.game_reports[p1.name]["action_history"]
        p2_actions = self.game_reports[p2.name]["action_history"]
        p1_utils = self.game_reports[p1.name]["util_history"]
        p2_utils = self.game_reports[p2.name]["util_history"]

        p1_action_counts = {action: p1_actions.count(action) for action in self.valid_actions}
        p2_action_counts = {action: p2_actions.count(action) for action in self.valid_actions}
        p1_invalid_moves = len([a for a in p1_actions if a not in self.valid_actions])
        p2_invalid_moves = len([a for a in p2_actions if a not in self.valid_actions])
        p1_total_utility = sum(p1_utils)
        p2_total_utility = sum(p2_utils)

        return (
            f"Game {self.game_num} Summary ({p1.name} VS {p2.name}):\n"
            f"{p1.name}: Played {p1_action_counts.get(0, 0)} Rocks, {p1_action_counts.get(1, 0)} Papers, "
            f"{p1_action_counts.get(2, 0)} Scissors; Invalid Moves: {p1_invalid_moves}; "
            f"Final Utility: {p1_total_utility}.\n"
            f"{p2.name}: Played {p2_action_counts.get(0, 0)} Rocks, {p2_action_counts.get(1, 0)} Papers, "
            f"{p2_action_counts.get(2, 0)} Scissors; Invalid Moves: {p2_invalid_moves}; "
            f"Final Utility: {p2_total_utility}.\n"
        )

    def _save_detailed_game_report(self, p1, p2):
        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        filename = f"{timestamp}_game_{self.game_num}.json"
        report_path = os.path.join(self.detailed_reports_path, filename)

        report = {
            "game_number": self.game_num,
            "players": [p1.name, p2.name],
            "p1_history": self.game_reports[p1.name],
            "p2_history": self.game_reports[p2.name],
        }

        with open(report_path, "w") as f:
            json.dump(report, f, indent=4)

        self._log_or_print(f"Saved detailed report for Game {self.game_num} at {report_path}")
        

    def _log_message(self, message):
        """
        Log a message to the logging path if provided.
        """
        if self.logging_path:
            with open(self.logging_path, "a") as log_file:
                log_file.write(message + "\n")

    def _log_or_print(self, message):
        """
        Log or print the summary or final results.
        """
        if self.logging_path:
            self._log_message(message)
        else:
            print(message)



from agt_server.agents.base_agents.lemonade_agent import LemonadeAgent
from agt_server.local_games.lemonade_arena import LemonadeArena
import argparse
import random
import numpy as np


class QQQ(LemonadeAgent):
    def setup(self):
        self.state = None
        self.pos = None
        self.q_table = {(a, b): [10 for _ in range(12)]
                        for a in range(6) for b in range(6)}
        self.rel_mov_1 = []
        self.rel_mov_2 = []
        self.rel_actions = []
        self.lr = 0.9
        self.discount = 0.1
        self.explore = 0.5

    def test_consistent(self):
        opp1std = np.std(self.rel_mov_1)
        opp2std = np.std(self.rel_mov_2)
        return (opp1std <= 1, opp2std <= 1)

    def get_action(self):
       # TODO: Enter logic to pick next result here
        if self.state == None:
            self.pos = random.randint(0, 11)
        else:
            if random.random() < self.explore:
                action = random.randint(0, 11)
            else:
                action = np.argmax(self.q_table[self.state])
            # decrease explore chance
            self.explore *= 0.98
            self.pos = (self.pos + action) % 12
            self.rel_actions.append(action)
        return self.pos

    def update(self):
        opp1a = ((self.get_opp1_last_action() -
                 self.get_last_action()) % 12) // 2
        opp2a = ((self.get_opp2_last_action() -
                 self.get_last_action()) % 12) // 2
        if len(self.get_opp2_action_history()) > 1:
            self.rel_mov_1.append(
                (self.get_opp1_action_history()[-1] - self.get_opp1_action_history()[-2]) % 12)
            self.rel_mov_2.append(
                (self.get_opp1_action_history()[-1] - self.get_opp2_action_history()[-2]) % 12)
        new_state = (opp1a, opp2a)
        if self.state == None or len(self.rel_actions) == 0:
            self.state = new_state
            return
        last_action = self.rel_actions[-1]
        # make reward also somewhat value reducing opponent score
        utilities = self.calculate_utility(self.get_action_history(
        )[-1], self.get_opp1_action_history()[-1], self.get_opp2_action_history()[-1])
        new_reward = self.get_last_util() + (self.get_last_util() -
                                             max(utilities[1:])) / 2
        # update reward
        self.q_table[self.state][last_action] = (
            self.lr * (new_reward + self.discount *
                       np.max(self.q_table[new_state]))
            + (1 - self.lr) * self.q_table[self.state][last_action])
        # update state
        self.state = new_state

################### SUBMISSION #####################
agent_submission = QQQ("QQQ")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    # parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    if args.join_server:
        agent_submission.connect(ip=args.ip, port=args.port)
    else:
        arena = LemonadeArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent_submission,
                QQQ("Agent_1"),
                QQQ("Agent_2"),
                QQQ("Agent_3"),
                QQQ("Agent_4")
            ]
        )
        arena.run()



import random
import numpy as np
from uniform_policy import UniformPolicy
from agt_server.agents.base_agents.chicken_agent import ChickenAgent
import os


class QLearning(ChickenAgent):
    def __init__(self, name, num_possible_states, num_possible_actions, initial_state, learning_rate, discount_factor, exploration_rate, training_mode, save_path=None) -> None:
        self.num_possible_actions = num_possible_actions
        self.num_possible_states = num_possible_states
        self.learning_rate = learning_rate
        self.discount_factor = discount_factor
        self.exploration_rate = exploration_rate
        self.training_mode = training_mode
        self.save_path = save_path
        self.s = initial_state
        super().__init__(name)

    def setup(self):
        self.my_states = []
        self.training_policy = UniformPolicy(self.num_possible_actions)

        if self.save_path and os.path.isfile(self.save_path):
            with open(self.save_path, 'rb') as saved_q_table:
                self.q = np.load(saved_q_table)
                assert self.q.shape[0] == self.num_possible_states, "The Saved Q-Table has a different number of states than inputed, To train on the new states please delete the Saved Q-Table"
                assert self.q.shape[1] == self.num_possible_actions, "The number of possible actions in the saved file is different from the actual game, please delete and train again."
        else:
            # initialize Q to random [-1, 1]
            self.q = np.array([[random.uniform(-1, 1)
                                for _ in range(self.num_possible_actions)]
                               for _ in range(self.num_possible_states)])

        # begin with initial state and random action.
        self.a = self.training_policy.get_move(self.s)
        self.s_prime = None

    def choose_next_move(self, s_prime):
        # In the next round, your agent will be in state [s_prime]. What move will it play?
        # TODO: Fill in the exploration-exploitation algorithm, choosing your next action
        # If in training mode,
        #    - Randomly choose an action (with probability self.exploration_rate)
        #    - Choose the best action (with probability 1 - self.exploration_rate)
        # Otherwise
        #    - Always choose the best action because you are no longer training
        # Hint: np.argmax and self.training_policy.get_move() may be useful!
        if (self.training_mode and random.random() < self.exploration_rate):
            return self.training_policy.get_move(self.s)
        else:
            return np.argmax(self.q[s_prime])

    def get_action(self):
        return self.a

    def set_training_mode(self, training_mode):
        self.training_mode = training_mode

    def determine_state(self):
        # Determines the next state s_prime given the action histories and reward histories
        # Don't worry about implementing this for now, you will implement this later
        raise NotImplementedError

    def update_rule(self, reward):
        # TODO: fill in the Q-learning update algorithm, updating self.Q
        # Q(s, a) = Q(s, a) + α[r + γ max_{a'} Q(s', a') − Q(s, a)]
        # Hint: np.max and self.s_prime may be useful!
        self.q[self.s, self.a] += self.learning_rate * \
            (reward + self.discount_factor *
             np.max(self.q[self.s_prime]) - self.q[self.s, self.a])
        if self.save_path:
            with open(self.save_path, 'wb') as saved_q_table:
                np.save(saved_q_table, self.q)

    def update(self):
        self.s_prime = self.determine_state()
        my_last_util = self.get_last_util()
        self.update_rule(my_last_util)
        self.s = self.s_prime
        self.a = self.choose_next_move(self.s_prime)
        self.s_prime = None



from agt_server.agents.base_agents.lemonade_agent import LemonadeAgent
from agt_server.local_games.lemonade_arena import LemonadeArena
import argparse
import numpy as np


class Zenly(LemonadeAgent):
    def setup(self):
        self.state = 'sticky'
        self.tolerance = 5

    def transform(self, st):
        return 'fluid' if st == 'sticky' else 'sticky'

    def get_action(self):
        history1 = np.array([0] * 12)
        history2 = np.array([0] * 12)
        sight = min(40, len(self.get_action_history()))
        for i in range(sight):
            history1[self.get_opp1_action_history()[i]] += 1
            history2[self.get_opp2_action_history()[i]] += 1
        if max(history1) > sight / 3 and max(history2) > sight / 3:
            self.state = 'fluid'
            spot1 = np.random.choice(12, 1, p=history1 / np.sum(history1))
            spot1 = spot1[0]
            spot2 = np.random.choice(12, 1, p=history2 / np.sum(history2))
            spot2 = spot2[0]
            action = (np.random.choice(
                [spot1, spot2], 1, [0.5, 0.5])[0] + 6) % 12
            return action
        elif max(history1) <= sight / 4 and max(history2) <= sight / 4:
            return 5
        else:
            self.tolerance -= 1
            if self.tolerance < 0:
                self.state = self.transform(self.state)
                self.tolerance = 5
            if self.state == 'sticky':
                return 5
            if self.state == 'fluid':
                spot1 = np.random.choice(12, 1, p=history1 / np.sum(history1))
                spot1 = spot1[0]
                spot2 = np.random.choice(12, 1, p=history2 / np.sum(history2))
                spot2 = spot2[0]
                action = (np.random.choice(
                    [spot1, spot2], 1, [0.5, 0.5])[0] + 6) % 12
                return action

    def update(self):
        pass


################### SUBMISSION #####################
agent_submission = Zenly("Zenly")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    # parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()
    
    if args.join_server:
        agent_submission.connect(ip=args.ip, port=args.port)
    else:
        arena = LemonadeArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent_submission,
                Zenly("Agent_1"),
                Zenly("Agent_2"),
                Zenly("Agent_3"),
                Zenly("Agent_4")
            ]
        )
        arena.run()



import random
from agt_server.local_games.base import LocalArena
import numpy as np
import pandas as pd 
from itertools import product

class SAArena(LocalArena):
    def __init__(self, num_goods, good_names = None, valuation_type = "additive", bid_range = (0, 100), num_rounds=1000, players=[], timeout=1, handin=False, logging_path=None, save_path=None, kth_price = 1):
        super().__init__(num_rounds, players, timeout, handin, logging_path, save_path)
        self.num_goods = num_goods
        self.kth_price = kth_price

        if good_names is not None:
            if len(good_names) != num_goods:
                raise ValueError("Length of good_names must match num_goods")
            self._goods_to_index = {name: idx for idx, name in enumerate(good_names)}
            self.goods = set(good_names)
            self._index_to_goods = {idx: name for name, idx in self._goods_to_index.items()}
        else:
            self._goods_to_index = SAArena._name_goods(num_goods)
            self.goods = set(self._goods_to_index.keys())
            self._index_to_goods = {value: key for key, value in self._goods_to_index.items()}


        self.bid_range = bid_range 
        self.valuation_type = valuation_type
        self.game_name = "Simultaneous Auction"
        self.price_history = []
        self.winner_history = []

        assert len(self.players) >= 2, "Arena must have at least 2 players"

        for idx in range(len(self.players)):
            player = self.players[idx]
            self.game_reports[player.name] = {
                "valuation_history": [],
                "bid_history": [],
                "util_history": [],
                "index": idx,
                "global_timeout_count": 0
            }

        self.results = []
        self.game_num = 1

    def run(self):
        """
        Runs a single auction game in which all players participate.
        First, each agent is restarted, then the game runs for num_rounds rounds.
        Finally, the results are summarized.
        """
        for player in self.players:
            self.run_func_w_time(player.restart, self.timeout, player.name)

        self.run_game(self.players)
        
        results = self.summarize_results()
        return results

    def run_helper(self, group):
        for _ in range(self.num_rounds):
            self._log_or_print(f"Game {self.game_num}: Auction Round - Goods: {self.goods}")

            for player in group:
                valuations = {good: random.randint(self.bid_range[0], self.bid_range[1]) for good in self.goods}

                player.num_goods = self.num_goods
                player.goods = self.goods
                player._index_to_goods = self._index_to_goods  
                player._goods_to_index = self._goods_to_index
                player.valuation_type = self.valuation_type
                player.valuations = np.zeros(self.num_goods)
                for good, value in valuations.items():
                    player.valuations[self._goods_to_index[good]] = value
                    
                self.game_reports[player.name]['valuation_history'].append(valuations)

            bids = {}
            for player in group:
                bid = self.run_func_w_time(
                    lambda: player.get_action(),
                    self.timeout,
                    player.name,
                    None
                )
                self.game_reports[player.name]['bid_history'].append(bid)
                player.game_report.game_history['my_bid_history'].append(bid)
                bids[player.name] = bid

            allocation, payments = self.compute_auction_result(bids)
            self.announce_outcome(allocation, payments)
            
            for player in group:
                player.game_report.game_history['my_payment_history'].append(payments[player.name])
                player.game_report.game_history['price_history'] = self.price_history
                player.game_report.game_history['winner_history'] = self.winner_history

            for player in group:
                won_goods = [good for good, winner in allocation.items() if winner == player.name] if allocation is not None else []
                valuations = self.game_reports[player.name]['valuation_history'][-1]
                
                base_sum = sum(valuations.get(good, 0) for good in won_goods)
                total_payment = payments[player.name]
                
                vt = self.valuation_type
                n = len(won_goods)
                
                if vt == 'additive':
                    round_val = base_sum
                elif vt == 'complement':
                    round_val = base_sum * (1 + 0.05 * (n - 1)) if n > 0 else 0
                elif vt == 'substitute':
                    round_val = base_sum * (1 - 0.05 * (n - 1)) if n > 0 else 0
                elif vt == 'randomized':
                    multiplier = 1.0
                    for _ in range(n):
                        multiplier *= random.uniform(0.95, 1.05)
                    round_val = base_sum * multiplier
                else:
                    round_val = base_sum

                round_util = round_val - total_payment

                self.game_reports[player.name]['util_history'].append(round_util)
                player.game_report.game_history['my_utils_history'].append(round_util)


            for player in group:
                self.run_func_w_time(player.update, self.timeout, player.name)
            
            self.game_num += 1

    def compute_auction_result(self, bids):
        """
        Computes the auction outcome for each good individually.
        
        Each player's bid is assumed to be a dictionary mapping goods to bid values.
        For each good in self.goods:
        - Determine the highest bid (if any) among all players.
        - Award the good to the player with the highest bid.
        - That player pays the amount of their bid for that good.
        
        Returns:
            allocation: dict mapping each good to the winning player's name (or None if no valid bid).
            payments: dict mapping each player's name to their total payment (sum of bids on goods they won).
        """
        allocation = {}
        # Initialize payments for each player by their address.
        payments = {player.name: 0 for player in self.players}
        prices = {}

        for good in self.goods:
            bid_tuples = []
            # Collect all valid bids for the good.
            for addr, bid_bundle in bids.items():
                if bid_bundle is not None:
                    bid_value = bid_bundle.get(good, None)
                    if bid_value is not None and bid_value > 0:
                        bid_tuples.append((bid_value, addr))
            
            if bid_tuples:
                # Sort bids in descending order by bid value.
                sorted_bids = sorted(bid_tuples, key=lambda x: x[0], reverse=True)
                winner = sorted_bids[0][1]
                # Determine kth highest bid price.
                kth_index = self.kth_price - 1  # since list indices start at 0
                if kth_index < len(sorted_bids):
                    kth_bid = sorted_bids[kth_index][0]
                else:
                    # If there are fewer than kth bids, use the lowest bid available.
                    kth_bid = sorted_bids[-1][0]
            else:
                winner = None
                kth_bid = None
            
            allocation[good] = winner
            prices[good] = kth_bid
            if winner is not None and kth_bid is not None:
                payments[winner] += kth_bid

        self.price_history.append(prices)
        self.winner_history.append(allocation)
        return allocation, payments

    def announce_outcome(self, allocation, payments):
        """
        Announces the auction outcome.
        """
        message = f"Auction outcome:\n Allocation: {allocation} \n Payments: {payments}"
        self._log_or_print(message)
        return message

    def run_game(self, group):
        """
        Runs a full auction game for a given trio of players.
        If in handin mode, attempts to load a shortcut result; otherwise,
        it executes the auction rounds via run_helper.
        Then, prints each agent's results and updates the overall results.
        """
        self.run_helper(group)

    def print_results(self, p):
        """
        Prints and returns the total utility for a given agent.
        """
        total_util = sum(self.game_reports[p.name].get('util_history', []))
        print(f"{p.name} got a total auction utility of {total_util}")
        return total_util

    def summarize_results(self):
        summary_data = []

        for player in self.players:
            player_name = player.name
            util_history = self.game_reports[player_name].get('util_history', [])

            total_util = sum(util_history)
            avg_util = total_util / len(util_history) if util_history else 0

            summary_data.append([player_name, total_util, avg_util])

        df = pd.DataFrame(summary_data, columns=["Player", "Total Utility", "Avg Utility Per Round"])
        df = df.sort_values(by="Total Utility", ascending=False)

        print(f"\nFinal Auction Results after {self.game_num - 1} Rounds:\n", df)
        
        return df

    @staticmethod
    def _generate_sequence(alphabet = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'):
        """
        Generates an infinite sequence of strings based on the provided alphabet.
        
        Parameters:
        - alphabet (str, optional): The alphabet to use for generating the sequence. Defaults to 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'.
        
        Yields:
        - str: The next string in the sequence.
        """
        yield from alphabet 
        size = 2 
        while True:
            for letters in product(alphabet, repeat=size):
                yield ''.join(letters)
            size += 1
        
    @staticmethod
    def _name_goods(size, alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"):
        """
        Names the goods in a multidimensional world of a given shape using an alphabet.
        
        Parameters:
        - size (int): The total number of goods to be named.
        - alphabet (str, optional): The alphabet used for naming goods.
        
        Returns:
        - dict: A mapping from string names to multidimensional indices.
        """
        alphabet_generator = SAArena._generate_sequence(alphabet)
        map_dict = {}
        
        for flat_index in range(size):
            letter = next(alphabet_generator)
            map_dict[letter] = flat_index
                        
        return map_dict